import time as cputiming
from abc import ABC
from abc import abstractmethod


class Solver(ABC): 
    """Abstract class with the **iterate** abstact method.  

    This class defines the interface of **iterate**.  Any child class
    od Solver class will contain all variables, controls and procedure
    for applying **iterate**.
    """
    @abstractmethod
    def iterate(self, problem, unknows):
        """
        Abstract interface for iterate procedure.

        Args:
            problem: Any class describing a problem.
            unknows: Any class describing a problem unknows.

        Returns:
            self: We return the solver to get statistics
            unknows: Updated solution
            ierr (int): Error flag. 
                ierr == 0 no error occured.
                ierr != 0 unknows may not be accurate.        
        """
        return self, unknows, ierr
    
class CycleControls():
    """Class containg controls and infos to apply iterative solver.
    
    This class contains the variable and reverse communicaiton method
    for applying iteratively a solver to find a solution of given a
    problem. Constructor procedure setting main controls (e.g.,
    maximum iterations number, verbosity) and setting to zero counter
    variables (e.g. number of iterations performed)
        
    Args:
        max_iter (int) : maximum iterations number
        max_restart (int): maximum restart number
        verbose (int): verbosity
    """
    def __init__(self,
                 max_iterations=1000,
                 max_restarts=10,
                 verbose=0):
        """
        Constructor of CycleControls setting main controls.

        Constructor procedure setting main controls
        (e.g., maximum iterations number, verbosity)
        and setting to zero counter variables
        (e.g. number of iterations performed)
        
        Args:
            max_iterations (int) : maximum iterations number
            max_restarts (int): maximum restart number
            verbose (int): verbosity level
        """

        """      
        Controls of iterative algorithm
        """
        #: int: Maximum iteration number
        self.max_iterations = max_iterations
        #: int: Maximum number of restart
        self.max_restarts = max_restarts
        #: int: Algorithm verbosity. verbose==0 silent
        #: verbose>0 more infos are printed
        self.verbose = verbose

        #: User comunication flag 
        #: 1->set inputs, 2->stop criteria 3->study
        self.flag = 0  
        #: Second user communication to request for solver controls
        #: 1: controls next update 2: controls after failed update
        self.request = 0
        
        #: int: State comunication flag
        self.ierr = 0
        #: int: Iterations counter
        self.iterations = 0
        #: int: Restart counter
        self.restarts= 0
 
        """      
        Statistics of iterative algorithm
        """
        #: float: Cpu conter
        self.cpu_time = 0
        
    def reverse_communication_solver(self, solver, problem, unknows):
        """ 
        Subroutine to run reverse communition approach 
        of iterative solver.
        
        Args:
            solver (Solver): Class with iterate method
            problem: Problem description
            unknows: Problem unknows

        Returns:
            self (CycleControls): Returning changed class.
                Counters and statistics are changed.
            solver (Solver): Class modified with statistics
                             of application.
            unknows: Updated solution.
        """
        
        if (self.flag == 0):
            # study system
            self.flag = 3
            self.request = 0
            self.ierr = 0
            
            return self, unknows, solver
            
        if (self.flag == 1):
            # user has settled problem inputs and controls
            # now we update
            self.ierr = 0
            
            # Update cycle. 
            # If it succees goes to the evaluation of
            # system varaition (flag ==4 ).
            # In case of failure, reset controls
            # and ask new problem inputs or ask the user to
            # reset controls (flag == 2 + ierr=-1 ).
        
            if (self.restarts == 0):
                if (self.verbose >= 1):
                    print(' ')
                    print('UPDATE ' + str(self.iterations+1))
                cpu_update = 0.0
            else:
                print('UPDATE ' +
                      str(self.iterations + 1) +
                      ' | RESTART = ' +
                      str(self.restarts))
            #
            # update unknows
            #
            start_time = cputiming.time()
            [unknows, ierr, solver] = solver.iterate(problem, unknows)
            cpu_update += cputiming.time() - start_time
        
            #
            # check for successfulls update
            #
            if (ierr == 0):
                # info for succesfull update
                self.iterations += 1
                self.cpu_time += cputiming.time() - start_time
                if (self.verbose >= 1):                    
                    if (self.restarts == 0):
                        print(
                            'UPDATE SUCCEED CPU = ' + 
                            '{:.2f}'.format(cpu_update)
                        )
                    else:
                        print(
                            'UPDATE SUCCEED ' +
                            str(self.restarts) +
                            ' RESTARTS CPU =' +
                            '{:.2f}'.format(cpu_update)
                        )
                        print(' ')
                #
                # Ask to the user to evalute if stop cycling 
                # 
                self.flag = 2
                self.request = 1
                self.ierr = 0 
            else:
                #
                # update failed
                #
                print('UPDATE FAILURE')
            
                # try one restart more
                self.restarts += 1
            
                # stop if number max restart update is passed 
                if (self.restarts >= self.max_restarts):
                    self.flag = -1  # breaking cycle
                    self.request = 0
                    self.ierr = ierr
                else:
                    # ask the user to reset controls and problem inputs
                    self.flag = 1
                    self.request = 2
                    self.ierr = ierr
            return self, unknows, solver
       
        if (self.flag == 2):
            # check if maximum number iterations was achieded 
            if (self.iterations >= self.max_iterations):
                self.flag = -1
                if (self.verbose >= 1):
                    print(
                        'Update Number exceed limits' + 
                        str(self.max_iterations)
                    )
                # break cycle
                return self, unknows, solver

            # Let the use study the system
            self.nrestarts = 0  # we reset the count of restart
            self.flag = 3
            self.ierr = 0
            return self, unknows, solver

        if (self.flag == 3):
            # ask user new problem inputs in order to update 
            self.flag = 1
            self.request = 1
            self.ierr = 0
            return self, unknows, solver

class ConstrainedSolver(ABC): 
    """
    Abstract procedure describing the solver syncronizing i.e.,
    action of the solver on the problem unknows so they fit 
    into potential problem constraints. 
    
    Args:
    problem : class describing problem setup
    unknows : class describing problem solution

    Returns:
    self    : solver class. We return the solver
              to read statistic.
    unknows : updated solution after one 
    ierr    : integer ==0 if no error occurred.
              Values different from 0 are used to
              identifity the error occured
    
    """
    @abstractmethod
    def syncronize(self, problem, unknows):
        return unknows, ierr, self  



import sys
import time as cputiming
from copy import deepcopy as cp



# leave these lines to work without installation
import os; sys.path.append(os.path.abspath(
'../src/')
    )
from rcis import Solver
from rcis import CycleControls

class Real:
    """
    Class describing a problem solution on real line
    """
    def __init__(self, x0):
        """
        Constructor setting initial solution
        """
        #: solution  
        self.x = x0

        #: current time  
        self.time = 0.0

        #: current iteration 
        self.iteration = 0

    def Save(self, file):
        """
        Procedure for saving to file

        Args:
            file (str) :: file path where save x solution
        """        
        f.open(file, 'w')  
        f.write(self.x)  
        f.close()  

# the r before comment allows to write Latex expression containg backslash
class Parabola:
    r"""
    Class describing a parabola.

    It contains the coeffients  a, b descringing the parabola
    :math: `y=a \frac{x^2}{2} + bx`
    
    Args:
        a (float): Parabola concavity (term :math:`a \frac{x^2}{2}`)
        b (float): Parabola linear term (term :math:`b x`)
    """    
    def __init__(self, a, b=0, lb=-1e30, up=1e30):
        """
        Constructor Parabola
        """

        #: a (float): Parabola quadratic coefficient :math:`a \frac{x^2}{2}`
        self.a = a
        #: b (float): Parabola linear coefficient :math: `b x~
        self.b = b
        #: lower_bound (float): Lower bound for solution
        self.lower_bound = lb
        #: upper_bound (float): Upper bound for solution
        self.upper_bound = up


class GradientDescentControls:
    """
    Class with gradient decent controls. 
    
    """
    def __init__(self,
                 step0=0.1,
                 verbose=0):
        """
        Control parameters of gradient descent solver
        """
        # step length
        self.step = step0

        # info solver application
        self.verbose = 0


class InfoDescent:
    def __init__(self):
        """
        Class to store info of the solver
        """
        self.cpu_gradient = 0.0

        
class ParabolaDescent(Solver): 
    """
    We extend the class "Solver"
    ovverriding the "syncronize" and the "iterate"
    procedure. 
    """
    def __init__(self, ctrl=None):
        """
        Initialize solver with passed controls (or default)
        and initialize structure to store info on solver application
        """
        # init controls
        if (ctrl is None):
            self.ctrl = GradientDescentControls(0.1)
        else:
            self.ctrl = cp(ctrl)
        # init infos
        self.infos = InfoDescent()
            
    def syncronize(self, problem, unknows):
        """
        Since the there are no constrain
        the first procedure does nothing.
        """

        # example of how the setup of our problem
        # can influence solver execution
        if (
                unknows.x >= problem.lower_bound and
                unknows.x <= problem.upper_bound
        ):
            ierr = 0
        else:
            ierr = -1
            
        return unknows, ierr, self

    def iterate(self, problem, unknows):
        """
        The update is one step of gradient descent.
        Currently with explicit euler.
        """
        start_time = cputiming.time() 
        gradient_direction = -problem.a * unknows.x
        self.infos.cpu_gradient = cputiming.time() - start_time
        
        unknows.x += self.ctrl.step * gradient_direction
        unknows.time += self.ctrl.step
        
        # example of how the setup of our problem
        # can influence solver execution
        if (
                unknows.x >= problem.lower_bound and
                unknows.x <= problem.upper_bound
        ):
            ierr = 0
        else:
            ierr = -1

        self.test = unknows.x
        
        return unknows, ierr, self


def test_main():
    # init solution container and copy it
    sol = Real(1)

    # init inputs data
    data = Parabola(0.5)

    # init solver
    ctrl = GradientDescentControls(0.1)
    grad_desc = ParabolaDescent(ctrl)

    # init update cycle controls
    flags = CycleControls(100, verbose=0)
    
    # Extra controls to tune Gradient Descent controls
    # with an increasening time step
    min_step = 0.01
    max_step = 2
    step_expansion = 1.05
    step_contraction = 2

    # list to store 
    sol_old = cp(sol)
    hystory = []
    while (flags.flag >= 0):
        """
        Call reverse communication.
        Then select action according to flag.flag and flag.info
        """
        flags, sol, grad_desc = flags.reverse_communication_solver(
            grad_desc,
            data,
            sol)
        
        # set problem inputs 
        if (flags.flag == 1):
            """
            Here user have to settle poblem inputs
            """
            #Nothing to do for this problem

            if (flags.request == 1):  
                """
                Here user have to set solver
                controls for next update
                """
                grad_desc.ctrl.step = max(min(grad_desc.ctrl.step * step_expansion,max_step),min_step)
                    

                # First we copy data before update
                sol_old.x = cp(sol.x)


            elif ( flags.request == 2):
                """
                Here user have to reset solver
                controls after update failure
                """
                grad_desc.ctrl.step = max(min(grad_desc.ctrl.step * step_expansion,max_step),min_step)
            
        if (flags.flag == 2):
            """
            Here the user evalutes if
            if convergence is achieved 
            and breaks the cycle setting 
            
            flags.flag = -1
            flags.info = 0 
            """
            var = abs(sol.x - sol_old.x) / grad_desc.ctrl.step
            
            if (var < 1e-4):
                flags.flag = -1
                flags.info = 0

                
        if (flags.flag == 3):
            """
            Here the user can study anything 
            combaining solver, problem, and unknows.
            User can save data at this point.
            """ 
            sol.iteration = cp(flags.iterations)
            hystory.append([sol.time, cp(grad_desc.infos)])
        

    assert abs(sol.x) < 1e-3
    return 0


if __name__ == "__main__":
    sys.exit(test_main())
